;; Generated package description from rcirc-menu.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "rcirc-menu" "1.1" "A menu of all your rcirc connections" 'nil :commit "e5cd88718503dd2d93a281ea62ee3236fdb716ec" :url "https://elpa.gnu.org/nongnu/rcirc-menu.html" :authors '(("Alex Schroeder" . "alex@gnu.org")) :maintainer '("Alex Schroeder" . "alex@gnu.org") :keywords '("comm"))
